﻿/// MF Calculator
/// Copyright (C) 2012 Célia Sales
/// 
/// This program is free software: you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation, either version 3 of the License, or
/// (at your option) any later version.
/// 
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
/// 
/// You should have received a copy of the GNU General Public License
/// along with this program.  If not, see http://www.gnu.org/licenses/.
/// 

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.IO;
using System.Text;

namespace MFC
{
    /// <summary>
    /// Type of calculation
    /// </summary>
    public enum CalculationType
    {
        Overall = 0,
        Score = 1,
        Frequency = 2
    }

    /// <summary>
    /// Main MFCalc class. How to use:
    /// 1. Instantiate main class, setting N factor and score/frequency parameters
    /// 2. Invoke public methods to get matrix values and average, accordingly to the specified type
    /// 
    /// Sample:
    ///     MFC.Calculator mfc = new MFC.Calculator("myFile.mfc", 10, 50);
    ///     
    ///     for (int l = 0; l < mfc.NumberOfPersons; l++)
    ///     {
    ///         for (int c = 0; c < mfc.NumberOfPersons; c++)
    ///         {
    ///             ...mfc.GetMFCValue(MFC.CalculationType.Overall, l, c)...
    ///         }
    ///     }
    ///     
    ///     ...to know the number of items in the matrix: mfc.NumberOfItems
    ///     ...to know the number of persons in the matrix: mfc.NumberOfPersons
    ///     ...to know the matrix value for line l, column c: mfc.GetMFCValue(MFC.CalculationType.Overall, l, c)
    ///     ...to get the average for the specified type: mfc.GetMFCAverage(MFC.CalculationType.Overall)
    /// 
    /// </summary>
    public class Calculator
    {
        #region " Declarations "

        // Fields
        private int _nFactor = 10;                      // These two fields controls matrix calculation
        private int _scoreVsFrequency = 50;             // Changing to any one causes matrix to be recomputed
        private List<Person> _persons = null;           // Collection of persons in the group

        // Internal variables
        private string filePath = "";                   // Input file path, Xml or Csv
        private int max = 0, min = 0;                   // Auxiliar max and min values
        private bool needsCalculation = true;           // Controls whether or not matrix needs to be computed

        private Hashtable overallValues = null;         // Holds values for overall similarity
        private Hashtable scoreValues = null;           // Holds values for score similarity
        private Hashtable frequencyValues = null;       // Holds values for frequency similarity

        // Variables used to compute averages, max, min and deviation
        private double overallAverage = 0, scoreAverage = 0, frequencyAverage = 0;
        private double overallMaximum = 0, scoreMaximum = 0, frequencyMaximum = 0;
        private double overallMinimum = -1, scoreMinimum = -1, frequencyMinimum = -1;
        private double overallDeviation = 0, scoreDeviation = 0, frequencyDeviation = 0;

        #endregion


        #region " Properties "

        public int NFactor
        {
            get { return _nFactor; }
            set
            {
                _nFactor = value;
                needsCalculation = true;
            }
        }

        public int ScoreVsFrequency
        {
            get { return _scoreVsFrequency; }
            set
            {
                _scoreVsFrequency = value;
                needsCalculation = true;
            }
        }

        public int NumberOfPersons
        {
            get
            {
                if (_persons == null)
                    return 0;
                else
                    return _persons.Count;
            }
        }

        public int NumberOfItems
        {
            get
            {
                if (_persons != null && _persons.Count > 0)
                    return _persons[0].Items.Count;
                else
                    return 0;
            }
        }

        public List<Person> Persons
        {
            get { return _persons; }
            set { _persons = value; }
        }

        #endregion


        #region " Constructors "

        /// <summary>
        /// Initialize the class using the specified file path and the specified parameters
        /// </summary>
        /// <param name="filePath"></param>
        /// <param name="nFactor"></param>
        /// <param name="score"></param>
        public Calculator(string filePath, int nFactor, int score)
        {
            this.filePath = filePath;
            _nFactor = nFactor;
            _scoreVsFrequency = score;

            ProcessDocument();
        }

        #endregion


        #region " Private methods "

        private void ProcessDocument()
        {
            _persons = new List<Person>();

            // Decide whether it is a Xml or a Csv file
            var extension = Path.GetExtension(filePath).ToLower();

            try
            {
                if (extension == ".mfc")
                    ProcessXmlDocument(XDocument.Load(filePath));
                else
                    ProcessCsvDocument(File.ReadAllLines(filePath, Encoding.UTF8));

                // Total number of cell with values
                int numberOfValues = GetNumOfValues(_persons.Count);

                // Initialize matrixs
                overallValues = new Hashtable(numberOfValues);
                scoreValues = new Hashtable(numberOfValues);
                frequencyValues = new Hashtable(numberOfValues);

                LoadMaxAndMin();
                ComputeMFC();
            }
            catch (Exception)
            {
                // Error loading Xml or Csv file
            }
        }

        private void ProcessCsvDocument(string[] csvContent)
        {
            int index = 0;
            Person person = null;
            bool isFirstLine = true;

            foreach (var line in csvContent)
            {
                // Check if first line has just header info
                // Each row having first column empty will be ignored
                //if (values.Length > 0 && values[0].Trim() == "")
                if (isFirstLine)
                {
                    isFirstLine = false;
                    continue;
                }

                var values = line.Split(";".ToCharArray());

                index = 0;
                person = new Person();

                foreach (var value in values)
                {
                    if (index == 0)
                        person.Name = value.Trim();
                    else
                    {
                        double finalValue = 0;
                        double.TryParse(value, out finalValue);

                        person.Items.Add(finalValue.ToString());
                        if (person.LargestItem < finalValue)
                            person.LargestItem = finalValue;
                    }

                    index++;
                }

                _persons.Add(person);
            }
        }

        private void ProcessXmlDocument(XDocument xmlContent)
        {
            int index = 0;
            Person person = null;

            var queryCase = from p in xmlContent.Elements("sav_file").Elements("case") select p;
            foreach (var recordCase in queryCase)
            {
                index = 0;
                person = new Person();

                var queryItem = from i in recordCase.Elements("var") select i;
                foreach (var recordItem in queryItem)
                {
                    if (index == 0)
                        person.Name = recordItem.Value.Trim();
                    else
                    {
                        string missing = "";
                        if (recordItem.Attribute("Missing") != null)
                            missing = recordItem.Attribute("Missing").Value;

                        double value = 0;
                        double.TryParse(recordItem.Value, out value);

                        double finalValue = GetFinalValue(missing, value);

                        person.Items.Add(finalValue.ToString());
                        if (person.LargestItem < value)
                            person.LargestItem = value;
                    }

                    index++;
                }

                _persons.Add(person);
            }
        }

        private double GetFinalValue(string missing, double value)
        {
            double retVal = 0;
            double num1 = 0, num2 = 0, num3 = 0;
            string[] list;

            if (missing == "")
                return value;

            list = missing.Split('-');
            switch (list.Length)
            {
                case 3:
                    Double.TryParse(list[0].ToString().Replace('.', ','), out num1);
                    Double.TryParse(list[1].ToString().Replace('.', ','), out num2);
                    Double.TryParse(list[2].ToString().Replace('.', ','), out num3);

                    if (value > num1 && value < num2)
                        retVal = num3;
                    else
                        if (value > num2 && value < num1)
                            retVal = num3;
                        else
                            retVal = value;
                    break;

                case 2:
                    Double.TryParse(list[0].ToString().Replace('.', ','), out num1);
                    list = list[1].Split(',');
                    Double.TryParse(list[0].ToString().Replace('.', ','), out num2);
                    Double.TryParse(list[1].ToString().Replace('.', ','), out num3);

                    if (value > num1 && value < num2)
                        retVal = num3;
                    else
                        if (value > num2 && value < num1)
                            retVal = num3;
                        else
                            retVal = value;
                    break;

                default:
                    break;
            }

            if (list.Length == 1)
            {
                list = missing.Split(',');
                switch (list.Length)
                {
                    case 1:
                        Double.TryParse(list[0].ToString().Replace('.', ','), out num1);

                        if (value == num1)
                            retVal = 0;
                        break;

                    case 2:
                        Double.TryParse(list[0].ToString().Replace('.', ','), out num1);
                        Double.TryParse(list[1].ToString().Replace('.', ','), out num2);

                        if (value == num1 || value == num2)
                            retVal = 0;
                        break;

                    case 3:
                        Double.TryParse(list[0].ToString().Replace('.', ','), out num1);
                        Double.TryParse(list[1].ToString().Replace('.', ','), out num2);
                        Double.TryParse(list[2].ToString().Replace('.', ','), out num3);

                        if (value == num1 || value == num2 || value == num3)
                            retVal = 0;
                        else
                            retVal = value;

                        break;

                    default:
                        break;
                }
            }

            return retVal;
        }

        private void LoadMaxAndMin()
        {
            int auxMax = 0, auxMin = 0;
            int auxNumber = 0;

            for (int i = 0; i < _persons.Count; i++)
            {
                for (int j = 0; j < _persons[i].Items.Count; j++)
                {
                    auxNumber = 0;
                    Int32.TryParse(_persons[i].Items[j].ToString(), out auxNumber);

                    if (auxNumber < auxMin)
                        auxMin = auxNumber;
                    if (auxNumber > auxMax)
                        auxMax = auxNumber;
                }
            }

            max = auxMax;
            min = auxMin;
        }

        private int CalculateJFM(string parameter, Person p1, Person p2)
        {
            // j: the number of “joint” items raised by both the father and the mother
            // f: the number of items raised by the father and not by the mother
            // m: the number of items raised by the mother and not by the father
            int retVal = 0;

            int auxParamP1 = 0;
            int auxParamP2 = 0;

            for (int i = 0; i < p1.Items.Count; i++)
            {
                auxParamP1 = 0;
                auxParamP2 = 0;

                Int32.TryParse(p1.Items[i].ToString(), out auxParamP1);
                Int32.TryParse(p2.Items[i].ToString(), out auxParamP2);

                if (parameter == "F")
                {
                    if ((Math.Abs(auxParamP1 - min) > 0.00001) && (Math.Abs(auxParamP2 - min) < 0.00001))
                        retVal++;
                }
                else
                    if (parameter == "J")
                    {
                        if ((Math.Abs(auxParamP1 - min) > 0.00001) && (Math.Abs(auxParamP2 - min) > 0.00001))
                            retVal++;
                    }
                    else
                        if ((Math.Abs(auxParamP1 - min) < 0.00001) && (Math.Abs(auxParamP2 - min) > 0.00001))
                            retVal++;
            }

            return retVal;
        }

        private double CalculateSimCatws(Person p1, Person p2, int j, int f, int m)
        {
            /*
             * sumwcom := j + f + m - sumwdif
             * simcatws := sumwcom / (j + f + m)
             */

            Double simcatws = 0;
            Double auxSimcatws = (j + f + m);

            if (auxSimcatws != 0)
                simcatws = (auxSimcatws - CalculateSumWDiff(p1, p2)) / auxSimcatws;
            else
                simcatws = 1;

            return simcatws;
        }

        private double CalculateSumWDiff(Person p1, Person p2)
        {
            double retVal = 0;

            /* 
             * sumwdif := 0
             * for i:=1 to mc do begin
             *  sumwdif := sumwdif + abs(wo[i]-wy[i]) / (max-min)
             * end
             */

            int auxDiffF = 0;
            int auxDiffM = 0;

            for (int i = 0; i < p1.Items.Count; i++)
            {
                auxDiffF = 0;
                auxDiffM = 0;

                Int32.TryParse(p1.Items[i].ToString(), out auxDiffF);
                Int32.TryParse(p2.Items[i].ToString(), out auxDiffM);

                if ((max - min) != 0)
                    retVal = retVal + (double)Math.Abs(auxDiffF - auxDiffM) / (max - min);
            }

            return retVal;
        }

        private double FrequencySimilarity(int j, int f, int m)
        {
            double retVal = 0;

            float retVal21 = 0;
            retVal21 = (float)j / NFactor;
            retVal21 = (float)Math.Sqrt((double)retVal21);

            double retVal22 = (1 - Math.Abs(Math.Sqrt(((float)f / NFactor)) - Math.Sqrt((float)m / NFactor)));
            retVal = (retVal21 + retVal22) / 2;

            return retVal;
        }

        private double CalculateMFSimilarity(int lineIndex, int columnIndex, out double scoreSimilarity, out double frequencySimilarity)
        {
            scoreSimilarity = 0;
            frequencySimilarity = 0;

            int frequencyVsScore = 100 - ScoreVsFrequency;

            int j = CalculateJFM("J", _persons[lineIndex], _persons[columnIndex]);
            int f = CalculateJFM("F", _persons[lineIndex], _persons[columnIndex]);
            int m = CalculateJFM("M", _persons[lineIndex], _persons[columnIndex]);

            scoreSimilarity = CalculateSimCatws(_persons[lineIndex], _persons[columnIndex], j, f, m);
            frequencySimilarity = FrequencySimilarity(j, f, m);

            return ((ScoreVsFrequency * 0.01 * scoreSimilarity) + (frequencyVsScore * 0.01 * frequencySimilarity));
        }

        private int GetNumOfValues(int n)
        {
            return n <= 1 ? 1 : n + GetNumOfValues(n - 1);
        }

        private bool ComputeMFC()
        {
            if (!needsCalculation)
                return true;

            if (overallValues == null || scoreValues == null || frequencyValues == null)
                return false;

            overallValues.Clear();
            scoreValues.Clear();
            frequencyValues.Clear();

            double overallSum = 0, scoreSum = 0, frequencySum = 0;

            for (int lineIndex = 0; lineIndex < _persons.Count; lineIndex++)
            {
                for (int columnIndex = 0; columnIndex < _persons.Count; columnIndex++)
                {
                    // cells with values
                    if (lineIndex > columnIndex)
                    {
                        double score = 0, frequency = 0;
                        double overall = CalculateMFSimilarity(lineIndex, columnIndex, out score, out frequency);

                        string key = lineIndex.ToString() + ":" + columnIndex.ToString();

                        // save overall values
                        overallValues[key] = overall;
                        overallSum += overall;
                        overallMaximum = overall > overallMaximum ? overall : overallMaximum;
                        overallMinimum = overall < overallMinimum || overallMinimum == -1 ? overall : overallMinimum;

                        // save score values
                        scoreValues[key] = score;
                        scoreSum += score;
                        scoreMaximum = score > scoreMaximum ? score : scoreMaximum;
                        scoreMinimum = score < scoreMinimum || scoreMinimum == -1 ? score : scoreMinimum;

                        // save frequency values
                        frequencyValues[key] = frequency;
                        frequencySum += frequency;
                        frequencyMaximum = frequency > frequencyMaximum ? frequency : frequencyMaximum;
                        frequencyMinimum = frequency < frequencyMinimum || frequencyMinimum == -1 ? frequency : frequencyMinimum;
                    }
                }
            }

            int numOfValueForAvg = GetNumOfValues(_persons.Count - 1);

            // Averages
            overallAverage = overallSum / numOfValueForAvg;
            scoreAverage = scoreSum / numOfValueForAvg;
            frequencyAverage = frequencySum / numOfValueForAvg;

            // Standard deviations
            double overallVariation = 0;
            foreach (double value in overallValues.Values)
                overallVariation += Math.Pow(value - overallAverage, 2);
            overallDeviation = Math.Sqrt(overallVariation / numOfValueForAvg);

            double scoreVariation = 0;
            foreach (double value in scoreValues.Values)
                scoreVariation += Math.Pow(value - scoreAverage, 2);
            scoreDeviation = Math.Sqrt(scoreVariation / numOfValueForAvg);

            double frequecyVariation = 0;
            foreach (double value in frequencyValues.Values)
                frequecyVariation += Math.Pow(value - frequencyAverage, 2);
            frequencyDeviation = Math.Sqrt(frequecyVariation / numOfValueForAvg);

            needsCalculation = false;

            return true;
        }

        #endregion


        #region " Public methods "

        /// <summary>
        /// Returns the MFC matrix value for the specified calculation type
        /// </summary>
        /// <param name="type"></param>
        /// <param name="lineIndex"></param>
        /// <param name="columnIndex"></param>
        /// <returns></returns>
        public object GetMfcValue(CalculationType type, int lineIndex, int columnIndex)
        {
            if (!ComputeMFC())
                return null;

            object retVal = "";

            // First cell
            if (lineIndex == 0 && columnIndex == 0)
                retVal = "";
            else
                // first line: show names
                if (lineIndex == 0 && columnIndex > 0)
                    retVal = _persons[columnIndex - 1].Name;
                else
                    // first column: show names
                    if (lineIndex > 0 && columnIndex == 0)
                        retVal = _persons[lineIndex - 1].Name;
                    else
                        // cells with no values
                        if (columnIndex > 0 && columnIndex > lineIndex)
                            retVal = "";
                        else
                            // diagonal cells
                            if (lineIndex == columnIndex)
                                retVal = "1";
                            else
                            {
                                // cells with values
                                // line and column inputs correspond to presentation indexes, where first line and first column are person names,
                                // so we must convert them to hashtable indexes, which stores only cells with values
                                string key = (lineIndex - 1).ToString() + ":" + (columnIndex - 1).ToString();

                                switch (type)
                                {
                                    case CalculationType.Overall:
                                        retVal = overallValues[key];
                                        break;

                                    case CalculationType.Score:
                                        retVal = scoreValues[key];
                                        break;

                                    case CalculationType.Frequency:
                                        retVal = frequencyValues[key];
                                        break;
                                }
                            }

            return retVal;
        }

        /// <summary>
        /// Returns the MFC average for the specified calculation type
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public double GetMfcAverage(CalculationType type)
        {
            ComputeMFC();

            double retVal = 0;

            switch (type)
            {
                case CalculationType.Overall:
                    retVal = overallAverage;
                    break;

                case CalculationType.Score:
                    retVal = scoreAverage;
                    break;

                case CalculationType.Frequency:
                    retVal = frequencyAverage;
                    break;
            }

            return retVal;
        }

        /// <summary>
        /// Returns the MFC maximum value for the specified calculation type
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public double GetMfcMaximum(CalculationType type)
        {
            ComputeMFC();

            double retVal = 0;

            switch (type)
            {
                case CalculationType.Overall:
                    retVal = overallMaximum;
                    break;

                case CalculationType.Score:
                    retVal = scoreMaximum;
                    break;

                case CalculationType.Frequency:
                    retVal = frequencyMaximum;
                    break;
            }

            return retVal;
        }

        /// <summary>
        /// Returns the MFC minimum value for the specified calculation type
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public double GetMfcMinimum(CalculationType type)
        {
            ComputeMFC();

            double retVal = 0;

            switch (type)
            {
                case CalculationType.Overall:
                    retVal = overallMinimum;
                    break;

                case CalculationType.Score:
                    retVal = scoreMinimum;
                    break;

                case CalculationType.Frequency:
                    retVal = frequencyMinimum;
                    break;
            }

            return retVal;
        }

        /// <summary>
        /// Returns the MFC standard deviation value for the specified calculation type
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public double GetMfcDeviation(CalculationType type)
        {
            ComputeMFC();

            double retVal = 0;

            switch (type)
            {
                case CalculationType.Overall:
                    retVal = overallDeviation;
                    break;

                case CalculationType.Score:
                    retVal = scoreDeviation;
                    break;

                case CalculationType.Frequency:
                    retVal = frequencyDeviation;
                    break;
            }

            return retVal;
        }

        #endregion
    }
}
